# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/

import os.path

import pytest

REPO_ROOT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..")
TEST_DATA_DIR = os.path.join(REPO_ROOT_DIR, "tests", "test-data")


@pytest.fixture
def test_data_dir():
    return TEST_DATA_DIR


@pytest.fixture
def repo_root_dir():
    return REPO_ROOT_DIR


def pytest_addoption(parser):
    parser.addoption(
        "--update-copyright-notices",
        action="store_true",
        default=False,
        help="Overwrite expected values",
    )


@pytest.fixture
def update_copyright_notices(request):
    return request.config.getoption("--update-copyright-notices")



# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/

import os.path

import scmdata
import xarray as xr
from openscm_units import unit_registry

from mesmer_openscmrunner.draw_realisations import (
    _draw_realisations_from_mesmer_file_and_openscm_output,
)


def test_create_realisations_from_pre_run_data(test_data_dir):
    mesmer_bundle_file = os.path.join(
        test_data_dir, "mesmer-bundles", "test-generic-mesmer-bundle.pkl"
    )

    magicc_output_dir = os.path.join(test_data_dir, "magicc-rcmip-phase-1-output")
    magicc_output_gsat = os.path.join(
        test_data_dir,
        magicc_output_dir,
        "rcmip-phase-1_magicc7.1.0.beta-canesm5-r1i1p1f1_world_surface-air-temperature-change.csv",
    )
    magicc_output_ohc = os.path.join(
        test_data_dir,
        magicc_output_dir,
        "rcmip-phase-1_magicc7.1.0.beta-canesm5-r1i1p1f1_world_heat-content-ocean.csv",
    )

    run_years = range(1850, 2100 + 1)
    scenarios_to_run = ["ssp126"]

    openscm_gsat = scmdata.ScmRun(magicc_output_gsat).filter(
        scenario=scenarios_to_run, year=run_years
    )

    # convert OHC to hfds in a crude way
    openscm_ohc = scmdata.ScmRun(magicc_output_ohc).filter(scenario=scenarios_to_run)
    # TODO: check whether this should be earth surface area or ocean surface area to be compatible with what MESMER has done
    earth_surface_area = 510.1 * 10 ** 6 * unit_registry("km^2")
    openscm_hfds = (
        openscm_ohc.delta_per_delta_time(out_var="Heat Uptake|Ocean")
        .convert_unit("W")
        .filter(year=run_years)
    ) / earth_surface_area

    result = _draw_realisations_from_mesmer_file_and_openscm_output(
        mesmer_bundle_file=mesmer_bundle_file,
        openscm_gsat=openscm_gsat,
        openscm_hfds=openscm_hfds,
        n_realisations_per_scenario=30,
    )

    assert isinstance(result, xr.Dataset)
    assert set(result.data_vars) == {"tas"}
    assert set(result.dims) == {"realisation", "scenario", "z", "year"}

    assert set(result["scenario"].values) == set(
        openscm_gsat.get_unique_meta("scenario")
    )

    # make sure we can get onto a lat lon grid from what is saved
    result_reshaped = result.set_index(z=("lat", "lon")).unstack("z")
    assert set(result_reshaped.dims) == {
        "scenario",
        "realisation",
        "lon",
        "lat",
        "year",
    }


def test_create_realisations_from_pre_run_data_with_unrecognised_scenario_name(
    test_data_dir,
):
    mesmer_bundle_file = os.path.join(
        test_data_dir, "mesmer-bundles", "test-generic-mesmer-bundle.pkl"
    )

    magicc_output_dir = os.path.join(test_data_dir, "magicc-rcmip-phase-1-output")
    magicc_output_gsat = os.path.join(
        test_data_dir,
        magicc_output_dir,
        "rcmip-phase-1_magicc7.1.0.beta-canesm5-r1i1p1f1_world_surface-air-temperature-change.csv",
    )
    magicc_output_ohc = os.path.join(
        test_data_dir,
        magicc_output_dir,
        "rcmip-phase-1_magicc7.1.0.beta-canesm5-r1i1p1f1_world_heat-content-ocean.csv",
    )

    run_years = range(1850, 2150 + 1)
    scenarios_to_run = ["ssp126", "esm-ssp126"]

    openscm_gsat = scmdata.ScmRun(magicc_output_gsat).filter(
        scenario=scenarios_to_run, year=run_years
    )

    # convert OHC to hfds in a crude way
    openscm_ohc = scmdata.ScmRun(magicc_output_ohc).filter(scenario=scenarios_to_run)
    # TODO: check whether this should be earth surface area or ocean surface area to be compatible with what MESMER has done
    earth_surface_area = 510.1 * 10 ** 6 * unit_registry("km^2")
    openscm_hfds = (
        openscm_ohc.delta_per_delta_time(out_var="Heat Uptake|Ocean")
        .convert_unit("W")
        .filter(year=run_years)
    ) / earth_surface_area

    result = _draw_realisations_from_mesmer_file_and_openscm_output(
        mesmer_bundle_file=mesmer_bundle_file,
        openscm_gsat=openscm_gsat,
        openscm_hfds=openscm_hfds,
        n_realisations_per_scenario=30,
    )

    assert isinstance(result, xr.Dataset)
    assert set(result.data_vars) == {"tas"}
    assert set(result.dims) == {"realisation", "scenario", "z", "year"}

    assert set(result["scenario"].values) == set(
        openscm_gsat.get_unique_meta("scenario")
    )

    # make sure we can get onto a lat lon grid from what is saved
    result_reshaped = result.set_index(z=("lat", "lon")).unstack("z")
    assert set(result_reshaped.dims) == {
        "scenario",
        "realisation",
        "lon",
        "lat",
        "year",
    }



# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/
import datetime
import glob
import os.path

import pytest

LICENSE_TEMPLATE = """# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021{} MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/"""


EXCLUDE = ["_version.py"]


def test_source_code_headers(repo_root_dir, update_copyright_notices, n_lines=None):
    n_lines = len(LICENSE_TEMPLATE.split("\n")) if n_lines is None else n_lines

    files_to_check = glob.glob(
        os.path.join(repo_root_dir, "tests", "**", "*.py"), recursive=True
    ) + glob.glob(os.path.join(repo_root_dir, "src", "**", "*.py"), recursive=True)

    files_to_check = [f for f in files_to_check if os.path.basename(f) not in EXCLUDE]

    now = datetime.datetime.now()
    current_year = now.year
    if current_year == 2021:
        license_template = LICENSE_TEMPLATE.format("")
    else:
        license_template = LICENSE_TEMPLATE.format("-{}".format(current_year))

    for f in files_to_check:
        with open(f) as fh:
            contents = fh.read()

        if not contents.startswith(license_template):
            if update_copyright_notices:
                contents = "\n".join(contents.split("\n")[n_lines:])
                contents = "{}\n\n{}".format(license_template, contents)
                with open(f, "w") as fh:
                    fh.write(contents)
            else:
                raise ValueError(
                    "Update headers in source files (failing file: {})".format(f)
                )

    if update_copyright_notices:
        pytest.skip("Updated copyright notices")



# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/

import joblib
import mesmer.create_emulations
import xarray as xr


def _draw_realisations_from_mesmer_file_and_openscm_output(
    mesmer_bundle_file,
    openscm_gsat,
    openscm_hfds=None,
    n_realisations_per_scenario=5,
):
    """
    Parameters
    ----------
    mesmer_bundle_file : str
        File in which the MESMER bundle is saved

    openscm_gsat : :obj:`scmdata.ScmRun`
        Global-mean surface air temperature change output to be used to drive MESMER's emulations

    openscm_hfds : [None, :obj:`scmdata.ScmRun`]
        If supplied, global-mean ocean heat uptake output to be used to drive MESMER's emulations

    n_realisations_per_scenario : int
        Number of realisations to draw for each scenario in ``openscm_gsat`` (and ``openscm_hfds`` if supplied)

    Returns
    -------
    :obj:`xr.Dataset`
        Emulations for each scenario

    Raises
    ------
    ValueError
        The scenarios in `openscm_gsat`` and ``openscm_hfds`` (if supplied) aren't the same

    ValueError
        [TODO talk to Lea about what to do here] The scenarios in ``openscm_gsat`` don't match those for which MESMER has been calibrated

    ValueError
        [TODO talk to Lea about what to do here] The scenarios in ``openscm_gsat`` span a longer time than MESMER has been calibrated for

    ValueError
        ``openscm_gsat`` contains variables other than ``"Surface Air Temperature Change"``

    ValueError
        ``openscm_hfds`` contains variables other than ``"Heat Uptake|Ocean"``
    """
    mesmer_bundle = joblib.load(mesmer_bundle_file)

    gsat_scenarios = set(openscm_gsat.get_unique_meta("scenario"))
    hfds_scenarios = set(openscm_hfds.get_unique_meta("scenario"))
    if gsat_scenarios != hfds_scenarios:
        raise ValueError(
            "gsat_scenarios: {}, hfds_scenarios: {}".format(
                gsat_scenarios, hfds_scenarios
            )
        )

    openscm_gsat_for_mesmer = _prepare_openscm_gsat(openscm_gsat)
    openscm_hfds_for_mesmer = _prepare_openscm_hfds(openscm_hfds)

    hard_coded_hist_years = range(1850, 2014 + 1)
    hard_coded_scen_years = range(2015, 3000 + 1)

    out = []
    for scen_gsat in openscm_gsat_for_mesmer.groupby("scenario"):
        scenario = scen_gsat.get_unique_meta("scenario", True)
        # if scenario not in mesmer_bundle["time"]:
        #     raise KeyError("No MESMER calibration available for: {}".format(scenario))

        hist_tas = _filter_and_assert_1d(scen_gsat.filter(year=hard_coded_hist_years))
        scen_tas = _filter_and_assert_1d(scen_gsat.filter(year=hard_coded_scen_years))

        openscm_hfds_for_mesmer_scen = openscm_hfds_for_mesmer.filter(scenario=scenario)
        hist_hfds = _filter_and_assert_1d(
            openscm_hfds_for_mesmer_scen.filter(year=hard_coded_hist_years)
        )
        scen_hfds = _filter_and_assert_1d(
            openscm_hfds_for_mesmer_scen.filter(year=hard_coded_scen_years)
        )

        # TODO: remove hard-coding and actually map up with what MESMER expects
        preds_lt_scenario = {}
        for predictor in mesmer_bundle["params_lt"]["preds"]:
            if predictor == "gttas":
                hist_vals = hist_tas
                scen_vals = scen_tas
            elif predictor == "gttas2":
                hist_vals = hist_tas ** 2
                scen_vals = scen_tas ** 2
            elif predictor == "gthfds":
                hist_vals = hist_hfds
                scen_vals = scen_hfds

            preds_lt_scenario[predictor] = {"hist": hist_vals, scenario: scen_vals}

        result_scenario = mesmer.create_emulations.make_realisations(
            preds_lt=preds_lt_scenario,
            params_lt=mesmer_bundle["params_lt"],
            params_lv=mesmer_bundle["params_lv"],
            params_gv_T=mesmer_bundle["params_gv_T"],
            n_realisations=n_realisations_per_scenario,
            seeds=mesmer_bundle["seeds"],  # TODO make this user customisable
            land_fractions=mesmer_bundle["land_fractions"],
            time={
                "hist": scen_gsat.filter(year=hard_coded_hist_years)["year"].values,
                scenario: scen_gsat.filter(year=hard_coded_scen_years)["year"].values,
            },
        )

        result_scenario = result_scenario.squeeze(
            dim="scenario", drop=True
        ).expand_dims({"scenario": [scenario]})
        out.append(result_scenario)

    out = xr.merge(out)

    return out


def _prepare_openscm_gsat(openscm_gsat):
    return openscm_gsat.convert_unit("K")


def _prepare_openscm_hfds(openscm_hfds):
    return openscm_hfds.convert_unit("W/m^2")


def _filter_and_assert_1d(scmrun, **kwargs):
    filtered = scmrun.filter(**kwargs).values.squeeze()
    if len(filtered.shape) != 1:
        raise ValueError(
            "Filters gave non-1D data: {}, {}".format(filtered.shape, kwargs)
        )

    return filtered



# MESMER-OpenSCM Runner, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 MESMER-OpenSCM Runner contributors, listed in AUTHORS, and ETH Zurich.
# Licensed under the GNU General Public License v3.0 or later; see LICENSE or https://www.gnu.org/licenses/



from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

import versioneer

PACKAGE_NAME = "mesmer-openscmrunner"
DESCRIPTION = "Coupling between MESMER and OpenSCM-Runner"
KEYWORDS = [
    "MESMER",
    "netCDF",
    "python",
    "climate",
    "atmosphere",
    "simple climate model",
    "reduced complexity climate model",
    "emulation",
    "earth system model emulation",
]

AUTHORS = [
    ("Zeb Nicholls", "zebedee.nicholls@climate-energy-college.org"),
    ("Lea Beusch", "lea.beusch@env.ethz.ch"),
    ("Mathias Hauser", "mathias.hauser@env.ethz.ch"),
]
URL = "https://github.com/MESMER-group/mesmer-openscmrunner"
PROJECT_URLS = {
    "Bug Reports": "https://github.com/MESMER-group/mesmer-openscmrunner/issues",
    # "Documentation": "https://mesmer-openscmruner.readthedocs.io/en/latest",
    "Source": "https://github.com/MESMER-group/mesmer-openscmrunner",
}
LICENSE = "GPLv3+"  # I think this is the right short-hand...
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
]

REQUIREMENTS = [
    # "mesmer",  # once released
    "openscm-runner>=0.6",
    "scmdata>=0.9,<0.10",  # whilst running notebooks with rcmip phase 2 output
    "xarray==0.17.0",  # whilst running notebooks with rcmip phase 2 output
]
REQUIREMENTS_TESTS = [
    "codecov",
    "nbval",
    "pytest",
    "pytest-cov",
    "pytest-xdist",
]
REQUIREMENTS_NOTEBOOKS = [
    "ipywidgets",
    "notebook",
    "seaborn",
]
REQUIREMENTS_DOCS = [
    "nbsphinx",
    "sphinx",
    "sphinx_rtd_theme",
    "sphinx-copybutton",
]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=38.6.0", "wheel>=0.31.0"]
requirements_dev = [
    *[
        "bandit",
        "black",
        "black-nb",
        "flake8",
        "isort",
        "mypy",
        "nbdime",
        "openscm-zenodo",
        "pydocstyle",
        "pylint",
    ],
    *REQUIREMENTS_DEPLOY,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_NOTEBOOKS,
    *REQUIREMENTS_TESTS,
]

REQUIREMENTS_EXTRAS = {
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": requirements_dev,
    "docs": REQUIREMENTS_DOCS,
    "notebooks": REQUIREMENTS_NOTEBOOKS,
    "tests": REQUIREMENTS_TESTS,
}

SOURCE_DIR = "src"

# no tests/docs in `src` so don't need exclude
PACKAGES = find_packages(SOURCE_DIR)
PACKAGE_DIR = {"": SOURCE_DIR}
PACKAGE_DATA = {}

README = "README.rst"

with open(README, "r") as readme_file:
    README_TEXT = readme_file.read()


class MesmerOpenSCMRunner(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": MesmerOpenSCMRunner})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author=", ".join([author[0] for author in AUTHORS]),
    author_email=", ".join([author[1] for author in AUTHORS]),
    url=URL,
    project_urls=PROJECT_URLS,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
)
